/* $Id: cbq_config.c,v 1.12 1999/12/28 11:56:46 kjc Exp kjc $ */
/* derived from rtap of rsvp and rsvp_config of sun and then merged */
/*
 * Copyright (c) Sun Microsystems, Inc. 1994-1998 All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the SMCC Technology
 *      Development Group at Sun Microsystems, Inc.
 *
 * 4. The name of the Sun Microsystems, Inc nor may not be used to endorse or
 *      promote products derived from this software without specific prior
 *      written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS
 * SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" 
 * without express or implied warranty of any kind. 
 *  
 * These notices must be retained in any copies of any part of this software.
 */
/*******************************************************************

  Copyright (c) 1996 by the University of Southern California
  All rights reserved.

  Permission to use, copy, modify, and distribute this software and its
  documentation in source and binary forms for any purpose and without
  fee is hereby granted, provided that both the above copyright notice
  and this permission notice appear in all copies. and that any
  documentation, advertising materials, and other materials related to
  such distribution and use acknowledge that the software was developed
  in part by the University of Southern California, Information
  Sciences Institute.  The name of the University may not be used to
  endorse or promote products derived from this software without
  specific prior written permission.

  THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about
  the suitability of this software for any purpose.  THIS SOFTWARE IS
  PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  Other copyrights might apply to parts of this software and are so
  noted when applicable.

********************************************************************/

#include <stddef.h>

#ifdef CBQDAEMON
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include <netdb.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#else
#include "rsvp_daemon.h"
#endif
#include <altq/altq.h>
#include <altq/altq_cbq.h>

#ifdef CBQDAEMON
#include "rsvp_cbq.h"
#include "if_nameindex.h"
#else
#include "tc_cbq.h"
#endif

#ifdef INCLUDE_SYSTEM_H
#include "system.h"
int printf(), strcasecmp();
#endif

#define show_help(op)  printf(cmd_tab[op].cmd_help)

/*
 * Forward & External Declarations
 */
static int	pfxcmp(char *, char *);
static int	next_word(char **, char *);

static int	cbq_do_cmd(int, char *);
static int	get_ifname(char **, char **);
static int	get_addr(char **, struct in_addr *, struct in_addr *);
static int	get_port(const char *, u_int16_t *);
static int	get_proto(const char *, int *);
static int	cbq_interface_cmd(char *);
static int	cbq_class_cmd(char *);
static int	cbq_filter_cmd(char *);
static int	cbq_ctl_cmd(char *);
static int	cbq_delete_cmd(char *);
#ifdef INET6
static int	cbq_filter6_cmd(char *);
static int	get_ip6addr(char **, struct in6_addr *, struct in6_addr *);
#endif
static void	filt_freeall(void);
static int	filt_save(struct flow_filter *, char *, char *, int);
static int	filt_check_relation(struct flow_filter *,
				    struct flow_filter *);
static int	filt_disjoint(struct flow_filter *, struct flow_filter *);
static int	filt_subset(struct flow_filter *, struct flow_filter *);

/*
 * Globals
 */
#define MAX_NFLWDS      64
#define MAX_T           64

int             TNO = 1;  	/* Current Thread number */

static char	if_names[MAX_T][IFNAMSIZ]; 
static int	line_no = 0;
static char	*configfile = "command input";
static struct if_nameindex *if_namelist = NULL;

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

enum op_codes {
        /* order must be same as entries cmd_tab[].cmd_op below!! */
        OP_HELP = 1, 	OP_QUIT, 
	OP_IFACE,	OP_CLASS,	OP_FILTER,
	OP_CBQ,		OP_DEL,		
#ifdef INET6
	OP_FILTER6,
#endif
        OP_NULL, 	OP_BUG
};

/*	Following table MUST match enum order of op_codes !
 */
struct cmds {
	char           *cmd_verb;
	int             cmd_op;
	char           *cmd_help;
}		cmd_tab[] = {

  	{ "?",		OP_HELP, 	"Commands are:\n" },
	{ "help",	OP_HELP, 	" help | ?\n" },
	{ "quit",	OP_QUIT, 	" quit\n" },
	{ "interface",	OP_IFACE,	" interface if_name [bandwidth bps] [sched_type]\n" },
	{ "class",	OP_CLASS,	" class cbq if_name class_name parent [borrow] [admission ctlload|none] [maxburst count] [minburst count] [packetsize bytes] [priority pri] [pbandwidth percent] [red|rio] [ecn] [flowvalve]\n" },
	{ "filter",	OP_FILTER,	" filter if_name class_name dst [netmask #] dport src [netmask #] sport proto [tos # [tosmask #] [gpi #] [dontwarn]\n" },
	{ "cbq",	OP_CBQ,		" cbq if_name {enable|disable disable}\n" },
	{ "delete",	OP_DEL,		" delete if_name class_name\n" },
#ifdef INET6
	{ "filter6",	OP_FILTER6,	" filter6 if_name class_name dst[/prefix] dport src[/prefix] sport proto [flowlabel #][tclass # [tclassmask #]][gpi #] [dontwarn]\n" },
#endif
	{ "bug",	OP_BUG,		" bug (On/Off)\n" },
	{ "",		OP_NULL,	"" } /* MUST BE LAST IN CMD TABLE */
};

/*
 * Read the config file to learn about tunnel vifs and non-default phyint
 * parameters.
 */
int
cbq_config()
{
	FILE		*f;
	int		rc = 1;
	extern char	*solconfigfile;

	if (if_namelist != NULL)
		if_freenameindex(if_namelist);
	if_namelist = if_nameindex();
		
	log(LOG_INFO, 0, "CBQ config file is %s\n", solconfigfile);

	f = fopen(solconfigfile, "r");
	if (f == NULL) {
		log(LOG_ERR, errno, "Can't open %s", solconfigfile, 0);
		return (-1);
	}
	line_no = 0;
	while (rc)
		rc = DoCommand(solconfigfile, f);

	(void) fclose(f);

	/* release the saved filters for consistency check */
	filt_freeall();  

	/* enable the interfaces */
	(void) cbq_enableall();

	return 0;
}

/*
 *	cbqd_cmds_init(): Initialization routine, called to start things off.
 */
void
cbq_cmds_init()
{
	int i;

	for (i = 0; i < MAX_T; i++)
		/* do initialize */
		if_names[i][0] = '\0';
}

/*
 *  Do_Command(): Top-level routine to read the next line from a given
 *	file and execute the command it contains.
 */
int
DoCommand(char *infile, FILE *infp)
{
	char	cmd_line[256], cmd_op[80];
	struct	cmds *cmdp;
	char	*cp;
	int	rc;

	if (fgets(cmd_line, sizeof(cmd_line), infp) == NULL)
		/* EOF */
		return(0);
	line_no++;

	/* check escaped newline */
	while ((cp = strrchr(cmd_line, '\\')) != NULL && cp[1] == '\n') {
		if (fgets(cp, &cmd_line[256] - cp, infp) != NULL)
			line_no++;
	}

	/* remove trailing NL */
	cp = cmd_line + strlen(cmd_line) - 1;
	if (*cp == '\n')
		*cp = '\0';
	else if (!feof(infp)) {
		printf("LINE %d > 255 CHARS: %s.\n", line_no, cmd_line);
		exit(1);
	}
	/*** printf("DoCommand: %s\n", cmd_line); ***/

	if (cmd_line[0] == '#') {	/* Comment, skip this line */
		return(1);
	}
	cp = cmd_line;
	if (!next_word(&cp, cmd_op))
		return(1);
	if (cmd_op[0] == 'T') {
		TNO = atoi(&cmd_op[1]);
		if (!next_word(&cp, cmd_op))
			return(1);
	}
	cmdp = cmd_tab;
	while ((cmdp->cmd_op != OP_NULL) && pfxcmp(cmd_op, cmdp->cmd_verb))
		cmdp++;

	if (cmdp->cmd_op == OP_NULL) {
		if (cmd_op[0])
			printf(" ?? %s\n", cmd_op);
		return(1);
	}
	rc = cbq_do_cmd(cmdp->cmd_op, cp);
	if (rc == 0) {
		if (infile) {
			/* error in the config file.  cleanup and exit. */
			log(LOG_ERR, 0, "Config failed. Exiting.\n");
			(void) cbq_destroyall();
			(void) fclose(infp);
			exit(1);
		}
		else {
			/* interactive mode */
			printf("error: usage :");
			show_help(cmdp->cmd_op);
		}
	}
	return(1);
}


/*
 * Prefix string comparison: Return 0 if s1 string is prefix of s2 string, 1
 * otherwise.
 */
static int 
pfxcmp(s1, s2)
	register char  *s1, *s2;
{
	while (*s1)
		if (*s1++ != *s2++)
			return (1);
	return (0);
}

/*
 * Skip leading blanks, then copy next word (delimited by blank or zero, but
 * no longer than 63 bytes) into buffer b, set scan pointer to following 
 * non-blank (or end of string), and return 1.  If there is no non-blank text,
 * set scan ptr to point to 0 byte and return 0.
 */
int 
next_word(char **cpp, char *b)
{
	char           *tp;
	int		L;

	*cpp += strspn(*cpp, " \t");
	if (**cpp == '\0' || **cpp == '\n' || **cpp == '#')
		return(0);

	tp = strpbrk(*cpp, " \t\n#");
	L = MIN((tp)?(tp-*cpp):strlen(*cpp), 63);
	strncpy(b, *cpp, L);
	*(b + L) = '\0';
	*cpp += L;
	*cpp += strspn(*cpp, " \t");
	return (1);
}

static int
cbq_do_cmd(int op, char *cmdbuf)
{
	int i, rval;
    
	switch (op) {
	case OP_HELP:
		for (i = 0; i < OP_NULL; i++)
			show_help(i);
		rval = 1;
		break;
	case OP_QUIT:
		cbq_destroyall();
		exit(0);
		break;
	case OP_IFACE:
		rval = cbq_interface_cmd(cmdbuf);
		break;
	case OP_CLASS:
		rval = cbq_class_cmd(cmdbuf);
		break;
	case OP_FILTER:
		rval = cbq_filter_cmd(cmdbuf);
		break;
	case OP_CBQ:
		rval = cbq_ctl_cmd(cmdbuf);
		break;
	case OP_DEL:
		rval = cbq_delete_cmd(cmdbuf);
		break;
#ifdef INET6
	case OP_FILTER6:
		rval = cbq_filter6_cmd(cmdbuf);
		break;
#endif
#ifdef CBQDAEMON
	case OP_BUG:
		if (debug & DEBUG_CBQ) {
			/* turn off verbose */
			l_debug = LOG_INFO;
			debug &= ~DEBUG_CBQ;
		}
		else {
			/* turn on verbose */
			l_debug = LOG_DEBUG;
			debug |= DEBUG_CBQ;
		}
		break;
#endif
	default:
		printf("cbq command %d not supported\n", op);
		rval = 0;
		break;
	}
	return(rval);
}

#define EQUAL(s1, s2)	(strcmp((char *)(s1), (char *)(s2)) == 0)

char *cur_ifname(void)
{
	return (if_names[TNO]);
}

static int
get_ifname(char **cpp, char **ifnamep)
{
	char w[128], *ocp;
	struct if_nameindex *ifnp;

	ocp = *cpp;
	if (next_word(&ocp, w) && if_namelist != NULL)
		for (ifnp = if_namelist; ifnp->if_name != NULL; ifnp++)
			if (strcmp(w, ifnp->if_name) == 0) {
				/* if_name found. advance the word pointer */
				*cpp = ocp; 
				strcpy(if_names[TNO], w);
				*ifnamep = if_names[TNO];
				return (1);
			}

	/* this is not interface name. use one in the context. */
	if (if_names[TNO][0] == 0)
		return (0);
	*ifnamep = if_names[TNO];
	return (1);
}

/* set address and netmask in network byte order */
static int
get_addr(char **cpp, struct in_addr *addr, struct in_addr *mask)
{
	char w[128], *ocp;
	u_int tmp;
	
	addr->s_addr = 0;
	mask->s_addr = 0xffffffff;

	if (!next_word(cpp, w))
		return (0);

	if ((tmp = inet_addr((char *)w)) == INADDR_NONE) {
		/* try gethostbyname */
		struct hostent *h;

		if ((h = gethostbyname(w)) == NULL
		    || h->h_addrtype != AF_INET || h->h_length != 4)
			return (0);
		
		bcopy(h->h_addr, &tmp, h->h_length);
	}

	addr->s_addr = tmp;

	/* check if netmask option is present */
	ocp = *cpp;
	if (next_word(&ocp, w) && EQUAL(w, "netmask")) {
		if (!next_word(&ocp, w))
			return (0);
		
		if (inet_aton((char *)w, (struct in_addr *)&tmp) == 0)
			return (0);

		mask->s_addr = tmp;
		*cpp = ocp;	
		return (1);
	}
	/* no netmask option */
	return (1);
}

/* returns service number in network byte order */
static int
get_port(name, port_no)
	const char *name;
	u_int16_t *port_no;
{
	struct servent *s;
	u_int16_t num;
	
	if (isdigit(name[0])) {
		num = strtol(name, NULL, 0);
		*port_no = htons(num);
		return (1);
	}

	if ((s = getservbyname(name, 0)) == NULL)
		return (0);

	*port_no = (u_int16_t)s->s_port;
	return (1);
}

static int
get_proto(name, proto_no)
	const char *name;
	int *proto_no;
{
	struct protoent *p;
	
	if (isdigit(name[0])) {
		*proto_no = strtol(name, NULL, 0);
		return (1);
	}

	if ((p = getprotobyname(name)) == NULL)
		return (0);

	*proto_no = p->p_proto;
	return (1);
}


#define	DISCIPLINE_NONE		0
#define	DISCIPLINE_CBQ		1
#define DISCIPLINE_CSZ		2
#define DISCIPLINE_CTLLOAD	3

static int
cbq_interface_cmd(char *cmdbuf) 
{
	char	w[128], *cp = cmdbuf;
	char	*ifname;
	long   	bandwidth = 0;
	u_int	is_efficient = 0;
	int	is_sched = 1;
	int     discipline = DISCIPLINE_NONE;
	u_int	is_wrr = 1;	/* weighted round-robin is default */
    
#ifdef CBQDAEMON
	extern int Debug_mode;
	if (Debug_mode)
		is_sched = 0;
#endif
	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/*
	 * Look for "bandwidth" & scheduling discipline options.
	 */
	while (next_word(&cp, w)) {
		if (EQUAL(w, "bandwidth")) {
			char *ptr;
			
			next_word(&cp, w);
			ptr = NULL;
			bandwidth = strtol(w, &ptr, 0);
			if (ptr != NULL) {
				if (*ptr == 'K' || *ptr == 'k')
					bandwidth *= 1000;
				else if (*ptr == 'M' || *ptr == 'm')
					bandwidth *= 1000000;
			}
		} else if (EQUAL(w, "efficient")) {
			is_efficient = 1;
		} else if (EQUAL(w, "cbq")) {
			discipline = DISCIPLINE_CBQ;
		} else if (EQUAL(w, "cbq-wrr")) {
			discipline = DISCIPLINE_CBQ;
			is_wrr = 1;
		} else if (EQUAL(w, "cbq-prr")) {
			discipline = DISCIPLINE_CBQ;
			is_wrr = 0;
		} else if (EQUAL(w, "admission_only")) {
			discipline = DISCIPLINE_CBQ;
			is_sched = 0;
		} else if (EQUAL(w, "nosched")) {
			is_sched = 0;
		} else if (EQUAL(w, "csz")) {
			log(LOG_ERR, 0,
			    "'csz' not yet supported - %s, line %d\n",
			    configfile, line_no);
			return (0);
		} else {
			log(LOG_ERR, 0,
			    "Unknown keyword '%s' in %s, line %d\n",
			    w, configfile, line_no);
			return (0);
		}
	}
	if (discipline == DISCIPLINE_CBQ) {
		/* Add this to list of CBQ interfaces */
		if (cbq_add_if(ifname, bandwidth, is_sched, is_wrr,
			       is_efficient)) {
			log(LOG_ERR, 0,
			    "can't init cbq on interface '%s'\n", ifname);
			return (0);
		}
	}
	else if (discipline == DISCIPLINE_NONE)
		log(LOG_WARNING, 0, "No discipline (e.g. cbq) specified!\n");
	else {
		log(LOG_ERR, 0, "unknown discipline!\n");
		return (0);
	}
	return (1);
}

static int
cbq_class_cmd(char *cmdbuf) 
{
	char	w[128], *cp = cmdbuf;
	char 	*ifname, class_name[128], parent_name[128], borrow_name[128];
	u_char  *clname = class_name;
	u_char  *parent = NULL;
	u_char  *borrow = NULL;
	u_int   pri = 1;
	int	flags = 0;
	u_int	pbandwidth = 100;
	u_int	exactbandwidth = 0;
	u_int	maxdelay = 0;		/* 0 means default */
	u_int	maxburst = 0;		/* 0 means default */
	u_int	minburst = 0;		/* 0 means default */
	u_int	av_pkt_size = 0;  /* 0 means use if mtu as default */
	u_int	max_pkt_size = 0; /* 0 means use if mtu as default */
	cbq_tos_t	admission_type = CBQ_QOS_NONE;
	cbq_ol_t	overlimit_action = OL_DELAY;
	int	ret;

	/* get scheduling class */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing scheduling discipline in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!EQUAL(w, "cbq")) {
		log(LOG_ERR, 0,
		    "unknown scheduling discipline '%s' in %s, line %d\n",
		    w, configfile, line_no);
		return (0);
	}

	/* get interface name */
	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get class name */
	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0, "missing class name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get parent name */
	if (!next_word(&cp, parent_name)) {
		log(LOG_ERR, 0, "missing parent class in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!EQUAL(parent_name, "null") && !EQUAL(parent_name, "NULL")) {
		parent = parent_name;
	}
	else {
		parent = NULL;
		flags |= CBQCLF_ROOTCLASS;
	}

	/* look for options. */
	while (next_word(&cp, w)) {
		if (EQUAL(w, "priority")) {
			next_word(&cp, w);
			pri = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "default")) {
			flags |= CBQCLF_DEFCLASS;
		} else if (EQUAL(w, "control")) {
			flags |= CBQCLF_CTLCLASS;
		} else if (EQUAL(w, "admission")) {
			next_word(&cp, w);
			if (EQUAL(w, "guaranteed"))
				admission_type = CBQ_QOS_GUARANTEED;
			else if (EQUAL(w, "predictive"))
				admission_type = CBQ_QOS_PREDICTIVE;
			else if (EQUAL(w, "cntlload"))
				admission_type = CBQ_QOS_CNTR_LOAD;
			else if (EQUAL(w, "cntldelay"))
				admission_type = CBQ_QOS_CNTR_DELAY;
			else if (EQUAL(w, "none"))
				admission_type = CBQ_QOS_NONE;
			else {
				log(LOG_ERR, 0,
				    "unknown admission type - %s, line %d\n",
				    w, line_no);
				return (0);
			}
		} else if (EQUAL(w, "maxdelay")) {
			next_word(&cp, w);
			maxdelay = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "borrow")) {
			borrow = parent;
#if 1
		/* support old style "borrow [parent]" */
		{
			char *tmp = cp;

			if (next_word(&tmp, borrow_name) &&
			    EQUAL(borrow_name, parent_name))
				/* old style, skip borrow_name */
				cp = tmp;
		}
#endif
		} else if (EQUAL(w, "pbandwidth")) {
			next_word(&cp, w);
			pbandwidth = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "exactbandwidth")) {
			char *ptr;
			
			next_word(&cp, w);
			ptr = NULL;
			exactbandwidth = (u_int)strtol(w, &ptr, 0);
			if (ptr != NULL) {
				if (*ptr == 'K' || *ptr == 'k')
					exactbandwidth *= 1000;
				else if (*ptr == 'M' || *ptr == 'm')
					exactbandwidth *= 1000000;
			}
			pbandwidth = 0; /* override pbandwidth */
		} else if (EQUAL(w, "maxburst")) {
			next_word(&cp, w);
			maxburst = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "minburst")) {
			next_word(&cp, w);
			minburst = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "packetsize")) {
			next_word(&cp, w);
			av_pkt_size = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "maxpacketsize")) {
			next_word(&cp, w);
			max_pkt_size = (u_int)strtol(w, NULL, 0);
		} else if (EQUAL(w, "red")) {
			flags |= CBQCLF_RED;
		} else if (EQUAL(w, "ecn")) {
			flags |= CBQCLF_ECN;
		} else if (EQUAL(w, "flowvalve")) {
			flags |= CBQCLF_FLOWVALVE;
		} else if (EQUAL(w, "rio")) {
			flags |= CBQCLF_RIO;
		} else {
			log(LOG_ERR, 0,
			    "Unknown keyword '%s' in %s, line %d\n",
			    w, configfile, line_no);
			return (0);	
		}
	}

	if ((flags & (CBQCLF_RED|CBQCLF_RIO)) == (CBQCLF_RED|CBQCLF_RIO)) {
		log(LOG_ERR, 0,
		    "both red and rio defined on interface '%s'\n",
		    ifname);
		return (0);
	}
	if ((flags & (CBQCLF_ECN|CBQCLF_FLOWVALVE))
	    && (flags & (CBQCLF_RED|CBQCLF_RIO)) == 0)
		flags |= CBQCLF_RED;
		
	ret = cbq_add_static_class(ifname, clname,
				   parent, borrow, pri, flags, pbandwidth,
				   maxdelay, maxburst, minburst,
				   av_pkt_size, max_pkt_size,
				   admission_type, overlimit_action,
				   exactbandwidth);
    
	if (ret) {
		log(LOG_ERR, 0, "can't add class '%s' on interface '%s'\n",
		    clname, ifname);
		return (0);
	}
    
	return (1);
}

static int
cbq_filter_cmd(char *cmdbuf) 
{
	char 	w[128], *cp = cmdbuf;
	char 	*ifname, class_name[128];
	struct flow_filter	sfilt;
	int	protocol;
	u_char	tos, tosmask;
	int	dontwarn = 0;
	int	ret;

	memset(&sfilt, 0, sizeof(sfilt));
	sfilt.ff_flow.fi_family = AF_INET;

	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0,
		    "missing class name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter destination Address */
	if (!get_addr(&cp, &sfilt.ff_flow.fi_dst, &sfilt.ff_mask.mask_dst)) {
		log(LOG_ERR, 0,
		    "bad filter destination address in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter destination port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0,
		    "missing filter destination port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!get_port(w, &sfilt.ff_flow.fi_dport)) {
		log(LOG_ERR, 0, "bad filter destination port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
    
	/* get filter source address */
	if (!get_addr(&cp, &sfilt.ff_flow.fi_src, &sfilt.ff_mask.mask_src)) {
		log(LOG_ERR, 0, "bad filter source address in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter source port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter source port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!get_port(w, &sfilt.ff_flow.fi_sport)) {
		log(LOG_ERR, 0, "bad filter source port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter protocol id */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter protocol id in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!get_proto(w, &protocol)) {
		log(LOG_ERR, 0, "bad protocol in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	sfilt.ff_flow.fi_proto = protocol; 

	while (next_word(&cp, w)) {
		if (EQUAL(w, "tos")) {
			tos = 0;
			tosmask = 0xff;
			
			if (next_word(&cp, w)) {
				tos = (u_char)strtol(w, NULL, 0);
				if (next_word(&cp, w)) {
					if (EQUAL(w, "tosmask")) {
						next_word(&cp, w);
						tosmask = (u_char)strtol(w, NULL, 0);
					}
				}
			}
			sfilt.ff_flow.fi_tos = tos; 
			sfilt.ff_mask.mask_tos = tosmask; 
		}
		else if (EQUAL(w, "gpi")) {
			if (next_word(&cp, w)) {
				sfilt.ff_flow.fi_gpi = strtol(w, NULL, 0);
				sfilt.ff_flow.fi_gpi =
					htonl(sfilt.ff_flow.fi_gpi);
			}
		}
		else if (EQUAL(w, "dontwarn"))
			dontwarn = 1;
	}
	
	/*
	 * Add the filter.
	 */
	ret = cbq_add_static_filter(ifname, class_name, &sfilt);
	if (ret) {
		log(LOG_ERR, 0,
		    "can't add filter to class '%s' on interface '%s'\n",
		    class_name, ifname);
		return (0);
	}

	/* save the filter for filter-rule consistency check */
	if (filt_save(&sfilt, ifname, class_name, dontwarn) < 0)
		/* fatal filter error! */
		return (0);

	return (1);
}

#ifdef INET6

#ifndef s6_addr32
#define s6_addr32 __u6_addr.__u6_addr32
#endif

static int
cbq_filter6_cmd(char *cmdbuf)
{
	char 	w[128], *cp = cmdbuf;
	char 	*ifname, class_name[128];
	struct flow_filter6	sfilt;
	int	protocol;
	u_char	tclass, tclassmask;
	int	dontwarn = 0;
	int	ret;

	memset(&sfilt, 0, sizeof(sfilt));
	sfilt.ff_flow6.fi6_family = AF_INET6;

	if (!get_ifname(&cp, &ifname)) {
		log(LOG_ERR, 0, "missing interface name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0, "missing class name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter destination address */
	if (!get_ip6addr(&cp, &sfilt.ff_flow6.fi6_dst,
			 &sfilt.ff_mask6.mask6_dst)) {
		log(LOG_ERR, 0, "bad destination address in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	
	/* get filter destination port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0,
		    "missing filter destination port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!get_port(w, &sfilt.ff_flow6.fi6_dport)) {
		log(LOG_ERR, 0, "bad filter destination port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
    
	/* get filter source address */
	if (!get_ip6addr(&cp, &sfilt.ff_flow6.fi6_src,
			 &sfilt.ff_mask6.mask6_src)) {
		log(LOG_ERR, 0, "bad source address in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter source port */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter source port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!get_port(w, &sfilt.ff_flow6.fi6_sport)) {
		log(LOG_ERR, 0, "bad filter source port in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	/* get filter protocol id */
	if (!next_word(&cp, w)) {
		log(LOG_ERR, 0, "missing filter protocol id in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	if (!get_proto(w, &protocol)) {
		log(LOG_ERR, 0, "bad protocol in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}
	sfilt.ff_flow6.fi6_proto = protocol; 

	while (next_word(&cp, w)) {
		if (EQUAL(w, "tclass")) {
			tclass = 0;
			tclassmask = 0xff;

			if (next_word(&cp, w)) {
				tclass = (u_char)strtol(w, NULL, 0);
				if (next_word(&cp, w)) {
					if (EQUAL(w, "tclassmask")) {
						next_word(&cp, w);
						tclassmask =
						    (u_char)strtol(w, NULL, 0);
					}
				}
			}
			sfilt.ff_flow6.fi6_tclass = tclass; 
			sfilt.ff_mask6.mask6_tclass = tclassmask; 
		}
		else if (EQUAL(w, "gpi")) {
			if (next_word(&cp, w)) {
				sfilt.ff_flow6.fi6_gpi = strtol(w, NULL, 0);
				sfilt.ff_flow6.fi6_gpi =
					htonl(sfilt.ff_flow6.fi6_gpi);
			}
		}
		else if (EQUAL(w, "flowlabel")) {
			if (next_word(&cp, w)) {
				sfilt.ff_flow6.fi6_flowlabel =
					strtol(w, NULL, 0) & 0x000fffff;
				sfilt.ff_flow6.fi6_flowlabel =
					htonl(sfilt.ff_flow6.fi6_flowlabel);
			}
		}
		else if (EQUAL(w, "dontwarn"))
			dontwarn = 1;
	}

	/*
	 * Add the filter.
	 */
	ret = cbq_add_static_filter(ifname, class_name,
				    (struct flow_filter *)&sfilt);
	if (ret) {
		log(LOG_ERR, 0,
		    "can't add filter to class '%s' on interface '%s'\n",
		    class_name, ifname);
		return (0);
	}

	/* save the filter for filter-rule consistency check */
	if (filt_save((struct flow_filter *)&sfilt,
		      ifname, class_name, dontwarn) < 0)
		/* fatal filter error! */
		return (0);

	return (1);
}

static int
get_ip6addr(char **cpp, struct in6_addr *addr, struct in6_addr *mask)
{
	char w[128], *prefix;
	u_char *cp;
	int len;

	*addr = in6addr_any;  /* set all 0 */
	*mask = in6addr_any;  /* set all 0 */

	if (!next_word(cpp, w))
		return (0);

	if (EQUAL(w, "0"))
		/* abbreviation of a wildcard (::0) */
		return (1);

	if ((prefix = strchr(w, '/')) != NULL) {
		/* address has prefix length */
		*prefix++ = '\0';
	}

	if (inet_pton(AF_INET6, w, addr) <= 0)
		return (0);

	if (IN6_IS_ADDR_UNSPECIFIED(addr) && prefix == NULL)
		/* wildcard */
		return (1);

	/* convert address prefix length to address mask */
	if (prefix != NULL) {
		len = strtol(prefix, NULL, 0);
		if ((len < 0) || (len > 128))
			return (0);
		for (cp = (u_char *)mask; len > 7; len -= 8)
			*cp++ = 0xff;
		if (len > 0)
			*cp = (0xff << (8 - len)) & 0xff;

		addr->s6_addr32[0] &= mask->s6_addr32[0];
		addr->s6_addr32[1] &= mask->s6_addr32[1];
		addr->s6_addr32[2] &= mask->s6_addr32[2];
		addr->s6_addr32[3] &= mask->s6_addr32[3];
	}
	else
		/* full mask */
		memset(mask, 0xff, sizeof(struct in6_addr));

	return (1);
}

#endif /* INET6 */

static int
cbq_ctl_cmd(char *cmdbuf)
{
	char	w[128], *cp = cmdbuf;
	char	*ifname;
	int	cbq_state;
	int	rval;
    
	if (!get_ifname(&cp, &ifname)) {
		printf("missing interface name in %s, line %d",
		       configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, w)) {
		cbq_state = cbq_is_enabled(ifname);
		printf("cbq %s on %s\n",
		       cbq_state ? "enabled" : "disabled", ifname);
		return (1);
	}
	
	if (EQUAL(w, "enable")) {
		rval = cbq_enable(ifname);
		printf("cbq %s on %s\n",
		       (rval == 0) ? "enabled" : "enable failed!", ifname);
	}
	else if (EQUAL(w, "disable")) {
		rval = cbq_disable(ifname);
		printf("cbq %s on %s\n",
		       (rval == 0) ? "disabled" : "disable failed!", ifname);
	}
	else if (EQUAL(w, "reload")) {
		printf("reinitializing cbq...\n");
		cbq_destroyall();
		cbq_init();
	}
	else
		return (0);
	return (1);
}


static int
cbq_delete_cmd(char *cmdbuf)
{
	char	*cp = cmdbuf;
	char	*ifname, class_name[128];
	int	ret;
    
	if (!get_ifname(&cp, &ifname)) {
		printf("missing interface name in %s, line %d",
		       configfile, line_no);
		return (0);
	}

	if (!next_word(&cp, class_name)) {
		log(LOG_ERR, 0,
		    "missing class name in %s, line %d\n",
		    configfile, line_no);
		return (0);
	}

	ret = cbq_delete_static_class(ifname, class_name);
	if (ret) {
		log(LOG_ERR, 0,
		    "can't delete class '%s' on interface '%s'\n",
		    class_name, ifname);
		return (0);
	}

	return (1);
}

/*
 * functions to check the filter-rules.
 * copies of filters are kept during the configuration.
 * when a new filter is added, we check the relation to the existing filters
 * and if some inconsistency is found, produce an error or a warning message.
 *
 * filter matching is performed from the head of the list.
 * let
 *    S: a set of packets that filter s matches
 *    T: a set of packets that filter t matches
 * filter relations are:
 *   disjoint: S ^ T = empty
 *   subset:   S <= T
 *   intersect: S ^ T = not empty
 *
 * a new filter is disjoint or subset of the existing filters --> ok
 * a new filter is superset of an existing filter --> order problem
 * a new filter intersect an existing filter --> warning
 *
 * port-intersect: a special case we don't make warning
 *      - intersection is only port numbers
 *	- one specifies src port and the other specifies dst port
 * there must be no packet with well-known port numbers in
 * both src and dst ports.  so this is ok.
 */

struct filtent {
	struct filtent *next;
	char	if_name[IFNAMSIZ];	
	char	class_name[128];	
	int	line_no;
	int	dontwarn;
	struct flow_filter filt;
};

static struct filtent *filter_list;

#define FILT_DISJOINT		1
#define FILT_SUBSET		2
#define FILT_SUPERSET		3
#define FILT_INTERSECT		4
#define FILT_PORTINTERSECT	5

/* free all the saved filters */
static void
filt_freeall(void)
{
	struct filtent *ent;

	while ((ent = filter_list) != NULL) {
		filter_list = ent->next;
		free(ent);
	}
}

static int
filt_save(filt, if_name, class_name, dontwarn)
	struct flow_filter *filt;
	char *if_name, *class_name;
	int dontwarn;
{
	struct filtent *fp, *ent;
	int relation;

	if ((fp = malloc(sizeof(struct filtent))) == NULL) {
		log(LOG_ERR, errno, "can't malloc!");
		return (-1);
	}
	strcpy(fp->if_name, if_name);
	strcpy(fp->class_name, class_name);
	fp->filt = *filt;
	fp->line_no = line_no;
	fp->dontwarn = dontwarn;

	/* check possible inconsistency */
	for (ent = filter_list; ent != NULL; ent = ent->next) {
		if (strcmp(fp->if_name, ent->if_name) != 0)
			/* different interface */
			continue;
		
		relation = filt_check_relation(&fp->filt, &ent->filt);
		switch (relation) {
		case FILT_SUBSET:
		case FILT_DISJOINT:
			/* OK */
			break;

		case FILT_SUPERSET:
			log(LOG_ERR, 0,
			    "filters for \"%s\" at line %d and for \"%s\" at line %d has order problem!\n", 
			    fp->class_name, fp->line_no,
			    ent->class_name, ent->line_no);
			return (-1);
			break;

		case FILT_PORTINTERSECT:
#if 0
			/* if we should be pedantic, go into INTERSECT case */
#else
			break;
#endif			
		case FILT_INTERSECT:
			/*
			 * if the intersecting two filters belogin to the
			 * same class, it's ok.
			 */
			if (strcmp(fp->class_name, ent->class_name) == 0)
				break;
			if (fp->dontwarn || ent->dontwarn)
				break;
			    
			log(LOG_WARNING, 0,
			    "warning: filter for \"%s\" at line %d could override filter for \"%s\" at line %d\n", 
			    fp->class_name, fp->line_no,
			    ent->class_name, ent->line_no);
			break;
		}
	}
	/* add the new filter to the head of the list */
	fp->next = filter_list;
	filter_list = fp;
	return (0);
}

static int
filt_check_relation(front, back)
	struct flow_filter *front, *back;
{
	int rval;
	
	if (front->ff_flow.fi_family != back->ff_flow.fi_family)
		return (FILT_DISJOINT);

	if (filt_disjoint(front, back))
		return (FILT_DISJOINT);

	if ((rval = filt_subset(front, back)) == 1)
		return (FILT_SUBSET);

	if (filt_subset(back, front) == 1)
		return (FILT_SUPERSET);

	if (rval == 2)
		return (FILT_PORTINTERSECT);
	
	return (FILT_INTERSECT);
}

static int
filt_disjoint(front, back)
	struct flow_filter *front, *back;
{
	u_int32_t mask;
	u_int8_t tosmask;
	
	if (front->ff_flow.fi_family == AF_INET) {
		if (front->ff_flow.fi_proto != 0 && back->ff_flow.fi_proto != 0
		    && front->ff_flow.fi_proto != back->ff_flow.fi_proto)
			return (1);
		if (front->ff_flow.fi_sport != 0 && back->ff_flow.fi_sport != 0
		    && front->ff_flow.fi_sport != back->ff_flow.fi_sport)
			return (1);
		if (front->ff_flow.fi_dport != 0 && back->ff_flow.fi_dport != 0
		    && front->ff_flow.fi_dport != back->ff_flow.fi_dport)
			return (1);
		if (front->ff_flow.fi_gpi != 0 && back->ff_flow.fi_gpi != 0
		    && front->ff_flow.fi_gpi != back->ff_flow.fi_gpi)
			return (1);
		if (front->ff_flow.fi_src.s_addr != 0 &&
		    back->ff_flow.fi_src.s_addr != 0) {
			mask = front->ff_mask.mask_src.s_addr &
				back->ff_mask.mask_src.s_addr;
			if ((front->ff_flow.fi_src.s_addr & mask) !=
			    (back->ff_flow.fi_src.s_addr & mask))
				return (1);
		}
		if (front->ff_flow.fi_dst.s_addr != 0 &&
		    back->ff_flow.fi_dst.s_addr != 0) {
			mask = front->ff_mask.mask_dst.s_addr &
				back->ff_mask.mask_dst.s_addr;
			if ((front->ff_flow.fi_dst.s_addr & mask) !=
			    (back->ff_flow.fi_dst.s_addr & mask))
				return (1);
		}
		if (front->ff_flow.fi_tos != 0 && back->ff_flow.fi_tos != 0) {
			tosmask = front->ff_mask.mask_tos &
				back->ff_mask.mask_tos;
			if ((front->ff_flow.fi_tos & tosmask) !=
			    (back->ff_flow.fi_tos & tosmask))
				return (1);
		}
		return (0);
	}
#ifdef INET6
	else if (front->ff_flow.fi_family == AF_INET6) {
		struct flow_filter6 *front6, *back6;
		int i;

		front6 = (struct flow_filter6 *)front;
		back6 = (struct flow_filter6 *)back;

		if (front6->ff_flow6.fi6_proto != 0 &&
		    back6->ff_flow6.fi6_proto != 0 &&
		    front6->ff_flow6.fi6_proto != back6->ff_flow6.fi6_proto)
			return (1);
		if (front6->ff_flow6.fi6_flowlabel != 0 &&
		    back6->ff_flow6.fi6_flowlabel != 0 &&
		    front6->ff_flow6.fi6_flowlabel !=
		    back6->ff_flow6.fi6_flowlabel)
			return (1);
		if (front6->ff_flow6.fi6_sport != 0 &&
		    back6->ff_flow6.fi6_sport != 0 &&
		    front6->ff_flow6.fi6_sport != back6->ff_flow6.fi6_sport)
			return (1);
		if (front6->ff_flow6.fi6_dport != 0 &&
		    back6->ff_flow6.fi6_dport != 0 &&
		    front6->ff_flow6.fi6_dport != back6->ff_flow6.fi6_dport)
			return (1);
		if (front6->ff_flow6.fi6_gpi != 0 &&
		    back6->ff_flow6.fi6_gpi != 0 &&
		    front6->ff_flow6.fi6_gpi != back6->ff_flow6.fi6_gpi)
			return (1);
		if (!IN6_IS_ADDR_UNSPECIFIED(&front6->ff_flow6.fi6_src) &&
		    !IN6_IS_ADDR_UNSPECIFIED(&back6->ff_flow6.fi6_src)) {
			for (i=0; i<4; i++) {
				mask = front6->ff_mask6.mask6_src.s6_addr32[i]
					& back6->ff_mask6.mask6_src.s6_addr32[i];
				if ((front6->ff_flow6.fi6_src.s6_addr32[i] & mask) !=
				    (back6->ff_flow6.fi6_src.s6_addr32[i] & mask))
					return (1);
			}
		}
		if (!IN6_IS_ADDR_UNSPECIFIED(&front6->ff_flow6.fi6_dst) &&
		    !IN6_IS_ADDR_UNSPECIFIED(&back6->ff_flow6.fi6_dst)) {
			for (i=0; i<4; i++) {
				mask = front6->ff_mask6.mask6_dst.s6_addr32[i]
					& back6->ff_mask6.mask6_dst.s6_addr32[i];
				if ((front6->ff_flow6.fi6_dst.s6_addr32[i] & mask) !=
				    (back6->ff_flow6.fi6_dst.s6_addr32[i] & mask))
				return (1);
			}
		}
		if (front6->ff_flow6.fi6_tclass != 0 &&
		    back6->ff_flow6.fi6_tclass != 0) {
			tosmask = front6->ff_mask6.mask6_tclass &
				back6->ff_mask6.mask6_tclass;
			if ((front6->ff_flow6.fi6_tclass & tosmask) !=
			    (back6->ff_flow6.fi6_tclass & tosmask))
				return (1);
		}
		return (0);
	}
#endif /* INET6 */
	return (0);
}

/*
 * check if "front" is a subset of "back".  assumes they are not disjoint
 * return value 0: not a subset
 *              1: subset
 *              2: subset except src & dst ports
 *		   (possible port-intersect)
 */
static int
filt_subset(front, back)
	struct flow_filter	*front, *back;
{
	u_int16_t srcport, dstport;
	
	if (front->ff_flow.fi_family == AF_INET) {
		if (front->ff_flow.fi_proto == 0 &&
		    back->ff_flow.fi_proto != 0)
			return (0);
		if (front->ff_flow.fi_gpi == 0 && back->ff_flow.fi_gpi != 0)
			return (0);
		if (front->ff_flow.fi_src.s_addr == 0) {
			if (back->ff_flow.fi_src.s_addr != 0)
				return (0);
		}
		else if (back->ff_flow.fi_src.s_addr != 0 &&
			 (~front->ff_mask.mask_src.s_addr &
			  back->ff_mask.mask_src.s_addr))
			return (0);
		if (front->ff_flow.fi_dst.s_addr == 0) {
			if (back->ff_flow.fi_dst.s_addr != 0)
				return (0);
		}
		else if (back->ff_flow.fi_dst.s_addr != 0 &&
			 (~front->ff_mask.mask_dst.s_addr &
			  back->ff_mask.mask_dst.s_addr))
			return (0);
		if (~front->ff_mask.mask_tos & back->ff_mask.mask_tos)
			return (0);

		if (front->ff_flow.fi_sport == 0 &&
		    back->ff_flow.fi_sport != 0) {
			srcport = ntohs(back->ff_flow.fi_sport);
			dstport = ntohs(front->ff_flow.fi_dport);
			if (dstport > 0 /* && dstport < 1024 */ &&
			    srcport > 0 /* && srcport < 1024 */)
				return (2);
			return (0);
		}
		if (front->ff_flow.fi_dport == 0 &&
		    back->ff_flow.fi_dport != 0) {
			dstport = ntohs(back->ff_flow.fi_dport);
			srcport = ntohs(front->ff_flow.fi_sport);
			if (srcport > 0 /* && srcport < 1024 */ &&
			    dstport > 0 /* && dstport < 1024 */)
				return (2);
			return (0);
		}
			
		return (1);
	}
#ifdef INET6
	else if (front->ff_flow.fi_family == AF_INET6) {
		struct flow_filter6 *front6, *back6;
		int i;
		
		front6 = (struct flow_filter6 *)front;
		back6 = (struct flow_filter6 *)back;

		if (front6->ff_flow6.fi6_proto == 0 &&
		    back6->ff_flow6.fi6_proto != 0)
			return (0);
		if (front6->ff_flow6.fi6_flowlabel == 0 &&
		    back6->ff_flow6.fi6_flowlabel != 0)
			return (0);
		if (front6->ff_flow6.fi6_gpi == 0 &&
		    back6->ff_flow6.fi6_gpi != 0)
			return (0);
		
		if (IN6_IS_ADDR_UNSPECIFIED(&front6->ff_flow6.fi6_src)) {
			if (!IN6_IS_ADDR_UNSPECIFIED(&back6->ff_flow6.fi6_src))
				return (0);
		}
		else if (!IN6_IS_ADDR_UNSPECIFIED(&back6->ff_flow6.fi6_src))
			for (i=0; i<4; i++)
				if (~front6->ff_mask6.mask6_src.s6_addr32[i] &
				    back6->ff_mask6.mask6_src.s6_addr32[i])
					return (0);
		if (IN6_IS_ADDR_UNSPECIFIED(&front6->ff_flow6.fi6_dst)) {
			if (!IN6_IS_ADDR_UNSPECIFIED(&back6->ff_flow6.fi6_dst))
				return (0);
		}
		else if (!IN6_IS_ADDR_UNSPECIFIED(&back6->ff_flow6.fi6_dst))
			for (i=0; i<4; i++)
				if (~front6->ff_mask6.mask6_dst.s6_addr32[i] &
				    back6->ff_mask6.mask6_dst.s6_addr32[i])
					return (0);

		if (~front6->ff_mask6.mask6_tclass &
		    back6->ff_mask6.mask6_tclass)
			return (0);

		if (front6->ff_flow6.fi6_sport == 0 &&
		    back6->ff_flow6.fi6_sport != 0) {
			srcport = ntohs(back6->ff_flow6.fi6_sport);
			dstport = ntohs(front6->ff_flow6.fi6_dport);
			if (dstport > 0 /* && dstport < 1024 */ &&
			    srcport > 0 /* && srcport < 1024 */)
				return (2);
			return (0);
		}
		if (front6->ff_flow6.fi6_dport == 0 &&
		    back6->ff_flow6.fi6_dport != 0) {
			dstport = ntohs(back6->ff_flow6.fi6_dport);
			srcport = ntohs(front6->ff_flow6.fi6_sport);
			if (srcport > 0 /* && srcport < 1024 */ &&
			    dstport > 0 /* && dstport < 1024 */)
				return (2);
			return (0);
		}
	}
#endif /* INET6 */
	return (1);
}


